(*
 * $Id: Parser.sml,v 1.4 2001/06/07 09:51:28 kuhlmann Exp $
 *
 * Naive Parser
 *
 * Authors: Marco Kuhlmann <mk@picoli.com>
 *
 *)

exception Parser

(* syntactic categories (this should be in a separate structure
 * such that the same parser can be used for different sets of
 * rules *)

datatype SCat = S | NP | PP | Q | VP | LEX of Lexicon.cat

(* rule type *)

type rule = {cat : SCat, subcat : SCat list}

(* syntax tree type (this should probably be in a separate
 * structure as well *)

datatype STree = N of Lexicon.tag | B of SCat * STree list

(* transform a string into a list of preterminal categories *)

fun totags lexicon s =
  let
    fun totag s = (case Lexicon.tag lexicon s of
		     SOME [t] => N(t) | _ => raise Parser)
    val isws = fn #" " => true | _ => false
  in
    map totag (String.tokens isws s)
  end

(* return the categories from the topmost level of a syntax
 * tree *)

val getcats = map (fn c => (case c of
			      (N(_,cat))    => LEX cat
			    | (B(cat,tree)) => cat))

(* for two lists l1 and l2, check if l1 is a prefix of l2 *)

fun isprefix nil     _       = true
  | isprefix (x::xs) nil     = false
  | isprefix (x::xs) (y::ys) = x=y andalso isprefix xs ys

(* This is the main parse function, which implements the naive
 * inference parser. *)

fun parse lexicon rules input =
  let

    (* Initialise the agenda and a bag to store the results *)

    val agenda = Stack.new()
    val bag = Bag.new()

    (* Check if a rule is applicable; if yes, add the matching
     * category to the agenda. *)

    fun inferit prefix suffix (rule:rule) =
      let
	val c = (#cat rule)
	val s = (#subcat rule)

        (* If the subcat list of the current rule is a prefix of
	 * the current suffix, the rule fires, and we can rewrite
	 * the original string of tokens into a tree with the lhs
	 * as a new node. *)

	val _ = if isprefix s (getcats suffix)
		  then
		    let
		      val xs = rev prefix
		      val ys = List.take(suffix,length s)
		      val zs = List.drop(suffix,length s)
		      val conclusion = xs@((B(c,ys))::zs)
		    in
		      Stack.push agenda conclusion
		    end
		else ()
      in

	(* Now we handle the processing of the rest of the
	 * input. (Note that the prefix has to be reversed, as
	 * we always cons in front of it. *)

	case suffix of [] => () | (h::t) => inferit (h::prefix) t rule

      end

    (* update pushs a new item into a beg if it does not already
     * exists in there. *)

    fun update b i =
      if not (List.exists (fn x => x=i) (Bag.contents b))
	then Bag.push b i
      else ()

    (* Infer is a wrapper for inferit: Given a list of categories
     * and a rule, infer everything than can be inferred. *)

    fun infer cats (rule:rule) = inferit nil cats rule

    (* Main processing loop. As long as the agenda is non-empty,
     * all possible rules are tried on its topmost entry. In
     * case that this is a single category, we have reached an
     * intermediate result and can push it into the bag. *)

    fun process () =
      if not (Stack.empty agenda) then
	let
	  val cats = Stack.pop agenda;
	  val _ = case cats of [cat] => update bag cat | _ => ()
	in
	  map (infer cats) rules ; process()
	end
      else ()
  in

    (* push the input on top of the agenda, generate the
     * inferential closure, and return the results. *)

    Stack.push agenda (totags lexicon input) ;
    process() ;
    Bag.contents bag

  end

(* prettyprinter *)

fun pp taglist =
  let
    fun pp (N(s,_))    = " " ^ s ^ " "
      | pp (B(cat,ts)) = "(" ^ (StringAux.smap pp ts) ^ ")"
  in
    map (fn tag => print ("# " ^ (pp tag) ^ "\n")) taglist ; ()
  end
