(*
 * $Id: automaton1.sml,v 1.5 2001/05/29 17:57:57 kuhlmann Exp $
 *
 * Finite Set Automata (Naive Implementation)
 *
 * Authors: Marco Kuhlmann <mk@picoli.com>
 *
 *)

structure Automaton :> AUTOMATON =
  struct

    (* set up state-transition network *)

    structure Network = STNetworkFn(structure State = IntOrder
				    structure Label = CharOrder)

    open Network

    (* auxiliary functions for strings *)

    open StringAux

    (* abstract data type *)

    type automaton = {depth : int,
		      start : S.set,
		      trans : T.set,
		      final : S.set}

    (* access functions *)

    fun depth (auto : automaton) = #depth auto

    fun initials (auto : automaton) = #start auto

    fun transitions (auto : automaton) = #trans auto

    fun finals (auto : automaton) = #final auto

    (* pretty printer *)

    fun pp {depth = n, start = s, trans = t, final = f} =
      let
	val i2s = Int.toString
	val c2s = Char.toString
	fun stostring s = "[" ^ i2s s ^ "]"
	fun ttostring t = "[" ^ i2s (source t) ^ "(" ^ c2s (label t) ^ ")" ^ i2s (target t) ^ "]"
      in
	print ("* depth: " ^ i2s(n) ^ "\n" ^
	       "* start: " ^ smap stostring (S.listItems s) ^ "\n" ^
	       "* trans: " ^ smap ttostring (T.listItems t) ^ "\n" ^
	       "* final: " ^ smap stostring (S.listItems f) ^ "\n")
      end

    (* the empty automaton *)
  
    val empty =
      {depth = 0,
       start = S.singleton 0,
       trans = T.empty,
       final = S.empty}

    (* the atomic automaton *)

    fun atom a =
      {depth = 1,
       start = S.singleton 0,
       trans = T.singleton (new 0 a 1),
       final = S.singleton 1}

    (* return those states in an automaton that one can reach from any state in
     * states by edges labeled with char *)

    fun nextstates auto states char =
      (targets o (by char) o (from states)) (transitions auto)

    (* return the set of states within an automaton that one can reach given a
     * certain input *)

    fun reach auto word =
      if isEmpty word
	then initials auto
      else nextstates auto (reach auto (prefix word)) (last word)

    (* return the set of final states within an automaton that one can reach
     * given a certain input *)

    fun acceptors auto word = S.intersection (finals auto) (reach auto word)

    (* returns true iff word is accepted by automaton *)

    fun accept auto word = (not o S.isEmpty) (acceptors auto word)

    (* returns the tag list for a given word *)

    fun tag auto word = S.listItems (acceptors auto word)

    (* -- naive implementation -- *)

    (* auxiliary functions *)

    fun liftState offset n = n+offset

    fun liftTransition offset (m,x,n) = (m+offset,x,n+offset)

    (* compute the union of two automata *)

    fun union auto1 auto2 =
      let
	val {depth = d1, start = s1, trans = t1, final = f1} = auto1
	val {depth = d2, start = s2, trans = t2, final = f2} = auto2
      in
	{depth = d1+d2+1,
	 start = S.union s1 (S.map (liftState (d1+1)) s2),
	 trans = T.union t1 (T.map (liftTransition (d1+1)) t2),
	 final = S.union f1 (S.map (liftState (d1+1)) f2)}
      end
    
    (* connect p with edges labeled a to states, i.e., return a set of
     * edges that start in p and reach every state in states by a *)

    fun connect p a states = S.foldl (fn (q,ts) => T.add ts (new p a q)) T.empty states

    (* compute the concatenation of two automata *)

    fun concat auto1 auto2 =
      let
	val {depth = d1, start = s1, trans = t1, final = f1} = auto1
	val {depth = d2, start = s2, trans = t2, final = f2} = auto2
	val s2' = S.map (liftState d1) s2
	val t2' = T.map (liftTransition d1) t2
	val old = T.filter (fn (p,a,q) => S.member f1 q) t1
	val new = T.foldl (fn ((p,a,q),e) => T.union (connect p a s2') e) T.empty old
      in
	{depth = d1+d2,
	 start = s1,
	 trans = T.union (T.difference t1 old) (T.union new t2'),
	 final = S.map (liftState d1) f2}
      end

    (* given a word, generate an automaton *)

    fun fromWord word =
      if isEmpty word
	then empty
      else concat (fromWord (prefix word)) (atom (last word)) 

    (* add a word to an automaton *)

    fun addWord auto word = union auto (fromWord word)

  end
