(*
 * Die Struktur Automaton stellt die Funktionalitt von endlichen Automaten
 * fr Lexika gem der Signatur AUTOMATON (in automaton-sig.sml) bereit.
 * Eure Aufgabe ist es, fehlende Code-Teile zu ergnzen.
 *)

structure Automaton :> AUTOMATON =
  struct

    (* Als erstes wird die Struktur Network erzeugt. Sie stellt die dem
     * Automaten zu Grunde liegende Infrastruktur eines Zustand-bergang-
     * Netzwerkes (State Transition Network) bereit. Zustnde implementieren
     * wir durch Integers, die Label von bergngen durch Characters. *)

    structure Network = STNetworkFn(structure State = IntOrder
				    structure Label = CharOrder)

    (* Weil wir die Funktionen von Network so oft brauchen, ffnen wir
     * die Struktur hier, d.h. machen alle Operationen, die wir sonst nur
     * durch N.foo zugreifen knnten, als foo direkt sichtbar. *)

    open Network

    (* Wir brauchen auch die Hilfsfunktionen fr Zeichenketten aus der
     * ersten Programmierbung. *)

    open StringAux

    (* Jetzt knnen wir den abstrakten Datentyp automaton einfhren. Wir
     * implementieren ihn als ein Verbund mit vier Feldern: depth ist die
     * Tiefe der lngsten Kette von Transitionen, der wir in dem Automaten
     * mglicherweise folgen knnen (brauchen wir spter fr buchhalterische
     * Zwecke), start ist die Menge der Startzustnde, trans die Menge der
     * bergnge (Transitionen), final die Menge der Endzustnde. Die
     * Bezeichner S.set und T.set referieren auf die Teilstrukturen S und T
     * aus der Network-Struktur; es gibt also zwei verschiedene set-Typen,
     * einen fr Zustnde und einen fr bergnge. *)

    type automaton = {depth : int,
		      start : S.set,
		      trans : T.set,
		      final : S.set}

    (* Zu einem abstrakten Datentyp gehren auch immer die entsprechenden
     * Zugriffsfunktionen. Falls wir die Implementierung des ADTs spter
     * ndern, knnen wir einfach die Implementierung der Zugriffsfunktionen
     * anpassen, ohne dass wir jede einzelne Stelle im Code ndern mssen,
     * an denen diese Zugriffsfunktionen benutzt werden. *)

    fun depth (auto : automaton) = #depth auto

    fun initials (auto : automaton) = #start auto

    fun transitions (auto : automaton) = #trans auto

    fun finals (auto : automaton) = #final auto

    (* Hier folgt jetzt ein "Pretty Printer", der einzig und allein dazu da
     * ist, dass ihr die Korrektheit eurer Funktionen berprfen knnt.
     * ber die Implementierung braucht ihr euch keine Sorgen zu machen. *)

    fun pp {depth = n, start = s, trans = t, final = f} =
      let
	val i2s = Int.toString
	val c2s = Char.toString
	fun stostring s = "[" ^ i2s s ^ "]"
	fun ttostring t = "[" ^ i2s (source t) ^ "(" ^ c2s (label t) ^ ")" ^ i2s (target t) ^ "]"
      in
	print ("* depth: " ^ i2s(n) ^ "\n" ^
	       "* start: " ^ smap stostring (S.listItems s) ^ "\n" ^
	       "* trans: " ^ smap ttostring (T.listItems t) ^ "\n" ^
	       "* final: " ^ smap stostring (S.listItems f) ^ "\n")
      end

    (* An dieser Stelle solltet ihr die Definitionen des leeren Automaten und
     * die des atomaren Automaten einfgen. Der leere Automat (empty) soll gar
     * keine Wrter akzeptieren, der atomare Automat (atom a) soll gerade den
     * Buchstaben a akzeptieren. *)

    (* val empty = ... *)

    (* fun atom a = ... *)

    (* nextstates soll einen Automaten (auto), eine Menge von Zustnden (states)
     * und einen Buchstaben (char) als Argumente nehmen und die Menge derjenigen
     * Zustnde in auto zurckliefern, die man von irgendeinem Zustand in states
     * erreichen kann, wenn man nur bergngen folgt, die das Label char haben. *)

    (* fun nextstates auto states char = ... *)

    (* reach nimmt einen Automaten (auto) und eine Zeichenkette (word) als
     * Eingabe und liefert die Menge derjenigen Zustnde zurck, die man im
     * Automaten auto erreichen kann, wenn word die Eingabe ist. (Ihr knnt
     * hier die Hilfsfunktionen fr Zeichenketten aus StringAux benutzen.) *)

    (* fun reach auto word = ... *)

    (* (acceptors auto word) soll diejenigen Zustnde von (reach auto word)
     * zurckgeben, die in auto Endzustnde sind. *)

    (* fun acceptors auto word = ... *)

    (* (accept auto word) soll genau dann true liefern, wenn das Wort word
     * durch den Automaten auto erkannt wird. *)

    (* fun accept auto word = ... *)

    (* (tag auto word) liefert die Liste der "Tags" eines Wortes zurck.
     * Wenn der Automat auto ein Wort akzeptiert, dann tut er dies in einem
     * oder mehreren Zustnden. Tags sind bei uns die Nummern dieser
     * Zustnde. Wir brauchen sie spter, um das Lexikon bauen zu knnen. *)

    fun tag auto word = S.listItems (acceptors auto word)

    (* Nun kommt der Hauptteil. Zunchst einige Hilfsprozeduren. *)

    (* (findTransition auto state char) liefert den Wert SOME t zurck,
     * wenn es in auto einen bergang von dem Zustand state ber eine
     * Kante mit dem Label char gibt, ansonsten NONE. *)

    fun findTransition auto state char =
      T.find (fn t => (source t) = state andalso (label t) = char) (transitions auto)

    (* addTransition fgt einem Automaten eine neue Kante t' hinzu und
     * setzt die Tiefe des Automaten um eins herauf. *)

    fun addTransition {depth = d,
		       start = s,
		       trans = t,
		       final = f} t' =
      {depth = d+1,
       start = s,
       trans = T.add t t',
       final = f}

    (* addFinal fgt einem Automaten einen neuen Endzustand f' hinzu. *)

    fun addFinal {depth = d,
		  start = s,
		  trans = t,
		  final = f} f' =
      {depth = d,
       start = s,
       trans = t,
       final = S.add f f'}

    (* Und nun zur zentralen Prozedur: inscribe. Sie nimmt einen Automaten
     * auto, einen Zustand state und ein Wort word als Eingabe und schreibt
     * das Wort ab state in den Automaten "hinein". Wenn ein Teilwort von
     * word bereits im Automaten drinsteht, werden dabei keine neuen
     * bergnge angelegt. An der ersten Stelle jedoch, wo das Wort word
     * von den schon im Automaten eingetragenen Wrtern abweicht, mssen
     * neue bergnge zum Automaten hinzugefgt werden. Rckgabewert von
     * inscribe ist ein gegenber auto evtl. erweiterter Automat, der das
     * Wort word erkennt. *)

    (* Fr die Implementierung dieser Funktion solltet ihr die Funktionen
     * aus StringAux, findTransition, addTransition sowie evtl. weitere
     * Funktionen zum Beispiel aus S oder T (in Network) verwenden. Ihr
     * knnt den Prettyprinter benutzen um zu testen, ob eure Funktion
     * korrekt arbeitet. *)

    (* fun inscribe auto state word = ... *)

    (* Die letzten beiden Funktionen sind Einzeiler: (addWord auto word)
     * fgt dem Automaten auto das Wort word hinzu, (fromWord word)
     * generiert einen neuen Automaten, der gerade word akzeptiert. *)

    (* fun addWord auto word = ... *)

    (* fun fromWord word = ... *)

  end
