(*
 * $Id: automaton2.sml,v 1.5 2001/05/29 17:57:57 kuhlmann Exp $
 *
 * Finite Set Automata (Optimised Implementation)
 *
 * Authors: Marco Kuhlmann <mk@picoli.com>
 *
 *)

structure Automaton :> AUTOMATON =
  struct

    (* set up state-transition network *)

    structure Network = STNetworkFn(structure State = IntOrder
				    structure Label = CharOrder)

    open Network

    (* auxiliary functions for strings *)

    open StringAux

    (* abstract data type *)

    type automaton = {depth : int,
		      start : S.set,
		      trans : T.set,
		      final : S.set}

    (* access functions *)

    fun depth (auto : automaton) = #depth auto

    fun initials (auto : automaton) = #start auto

    fun transitions (auto : automaton) = #trans auto

    fun finals (auto : automaton) = #final auto

    (* pretty printer *)

    fun pp {depth = n, start = s, trans = t, final = f} =
      let
	val i2s = Int.toString
	val c2s = Char.toString
	fun stostring s = "[" ^ i2s s ^ "]"
	fun ttostring t = "[" ^ i2s (source t) ^ "(" ^ c2s (label t) ^ ")" ^ i2s (target t) ^ "]"
      in
	print ("* depth: " ^ i2s(n) ^ "\n" ^
	       "* start: " ^ smap stostring (S.listItems s) ^ "\n" ^
	       "* trans: " ^ smap ttostring (T.listItems t) ^ "\n" ^
	       "* final: " ^ smap stostring (S.listItems f) ^ "\n")
      end

    (* the empty automaton *)
  
    val empty =
      {depth = 0,
       start = S.singleton 0,
       trans = T.empty,
       final = S.empty}

    (* the atomic automaton *)

    fun atom a =
      {depth = 1,
       start = S.singleton 0,
       trans = T.singleton (new 0 a 1),
       final = S.singleton 1}

    (* return those states in an automaton that one can reach from any state in
     * states by edges labeled with char *)

    fun nextstates auto states char =
      (targets o (by char) o (from states)) (transitions auto)

    (* return the set of states within an automaton that one can reach given a
     * certain input *)

    fun reach auto word =
      if isEmpty word
	then initials auto
      else nextstates auto (reach auto (prefix word)) (last word)

    (* return the set of final states within an automaton that one can reach
     * given a certain input *)

    fun acceptors auto word = S.intersection (finals auto) (reach auto word)

    (* returns true iff word is accepted by automaton *)

    fun accept auto word = (not o S.isEmpty) (acceptors auto word)

    (* returns the tag list for a given word *)

    fun tag auto word = S.listItems (acceptors auto word)

    (* -- optimised implementation -- *)

    (* return SOME transition if in auto there is a transition from state to
     * some other state via char, NONE otherwise *)

    fun findTransition auto state char =
      T.find (fn t => (source t) = state andalso (label t) = char) (transitions auto)

    (* add a transition to an automaton and increase its depth *)

    fun addTransition {depth = d,
		       start = s,
		       trans = t,
		       final = f} t' =
      {depth = d+1,
       start = s,
       trans = T.add t t',
       final = f}

    (* add a new final state to an automaton *)

    fun addFinal {depth = d,
		  start = s,
		  trans = t,
		  final = f} f' =
      {depth = d,
       start = s,
       trans = t,
       final = S.add f f'}

    (* inscribe a word into an automaton *)

    fun inscribe auto state word =
      if isEmpty word
	then addFinal auto state
      else
	let
	  val fst = first word
	  val suf = suffix word
	in
	  (case findTransition auto state fst of
	     SOME t => inscribe auto (target t) suf
	   | NONE   => let
			 val t' = new state fst ((depth auto)+1)
		       in
			 inscribe (addTransition auto t') (target t') suf
		       end)
	end

    (* add a word to an automaton *)

    fun addWord auto word = inscribe auto 0 word

    (* given a word, generate an automaton *)

    fun fromWord word = addWord empty word

  end
